config_object_template = '/******************************************************************************\n* Copyright (c) 2017 - 2022 Xilinx, Inc.  All rights reserved.\n* Copyright (c) 2022 - 2024 Advanced Micro Devices, Inc. All Rights Reserved.\n* SPDX-License-Identifier: MIT\n******************************************************************************/\n\n#include "xil_types.h"\n#include "pm_defs.h"\n#include "pm_cfg_obj.h"\n\n#define PM_CONFIG_MASTER_SECTION_ID	0x101U\n#define PM_CONFIG_SLAVE_SECTION_ID	0x102U\n#define PM_CONFIG_PREALLOC_SECTION_ID	0x103U\n#define PM_CONFIG_POWER_SECTION_ID	0x104U\n#define PM_CONFIG_RESET_SECTION_ID	0x105U\n#define PM_CONFIG_SHUTDOWN_SECTION_ID	0x106U\n#define PM_CONFIG_SET_CONFIG_SECTION_ID	0x107U\n#define PM_CONFIG_GPO_SECTION_ID	0x108U\n\n#define PM_SLAVE_FLAG_IS_SHAREABLE	0x1U\n#define PM_MASTER_USING_SLAVE_MASK	0x2U\n\n#define PM_CONFIG_GPO1_MIO_PIN_34_MAP	(1U << 10U)\n#define PM_CONFIG_GPO1_MIO_PIN_35_MAP	(1U << 11U)\n#define PM_CONFIG_GPO1_MIO_PIN_36_MAP	(1U << 12U)\n#define PM_CONFIG_GPO1_MIO_PIN_37_MAP	(1U << 13U)\n\n#define PM_CONFIG_GPO1_BIT_2_MASK	(1U << 2U)\n#define PM_CONFIG_GPO1_BIT_3_MASK	(1U << 3U)\n#define PM_CONFIG_GPO1_BIT_4_MASK	(1U << 4U)\n#define PM_CONFIG_GPO1_BIT_5_MASK	(1U << 5U)\n\n#define SUSPEND_TIMEOUT	0xFFFFFFFFU\n\n#define PM_CONFIG_OBJECT_TYPE_BASE	0x1U\n\n<<MASTER_IPI_MASK_DEF>>\n\n#if defined (__ICCARM__)\n#pragma language=save\n#pragma language=extended\n#endif\n#if defined (__GNUC__)\n    const u32 XPm_ConfigObject[] __attribute__((used, section(".sys_cfg_data"))) =\n#elif defined (__ICCARM__)\n#pragma location = ".sys_cfg_data"\n__root const u32 XPm_ConfigObject[] =\n#endif\n{\n	/**********************************************************************/\n	/* HEADER */\n	2,	/* Number of remaining words in the header */\n	8,	/* Number of sections included in config object */\n	PM_CONFIG_OBJECT_TYPE_BASE,	/* Type of config object as base */\n	/**********************************************************************/\n	/* MASTER SECTION */\n<<MASTER_SECTION_DATA>>\n	/**********************************************************************/\n	/* SLAVE SECTION */\n<<SLAVE_SECTION_DATA>>\n	/**********************************************************************/\n	/* PREALLOC SECTION */\n<<PREALLOC_SECTION_DATA>>\n	/**********************************************************************/\n	/* POWER SECTION */\n<<POWER_SECTION_DATA>>\n	/**********************************************************************/\n	/* RESET SECTION */\n<<RESET_SECTION_DATA>>\n	/**********************************************************************/\n	/* SET CONFIG SECTION */\n<<SET_CONFIG_SECTION_DATA>>\n	/**********************************************************************/\n	/* SHUTDOWN SECTION */\n<<SHUTDOWN_SECTION_DATA>>\n	/**********************************************************************/\n	/* GPO SECTION */\n<<GPO_SECTION_DATA>>\n};\n#if defined (__ICCARM__)\n#pragma language=restore\n#endif'
